import asyncio
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict

from acp_sdk.server.store.utils import Stringable


class StoreModel(BaseModel):
    model_config = ConfigDict(extra="allow")


T = TypeVar("T", bound=BaseModel)
U = TypeVar("U", bound=BaseModel)


class Store(Generic[T], ABC):
    @abstractmethod
    async def get(self, key: Stringable) -> T | None:
        pass

    @abstractmethod
    async def set(self, key: Stringable, value: T | None) -> None:
        pass

    @abstractmethod
    def watch(self, key: Stringable, *, ready: asyncio.Event | None = None) -> AsyncIterator[T | None]:
        pass

    def as_store(self, model: type[U], prefix: Stringable = "") -> "Store[U]":
        return StoreView(model=model, store=self, prefix=prefix)


class StoreView(Store[U], Generic[U]):
    def __init__(self, *, model: type[U], store: Store[T], prefix: Stringable = "") -> None:
        super().__init__()
        self._model = model
        self._store = store
        self._prefix = prefix

    async def get(self, key: Stringable) -> U | None:
        value = await self._store.get(self._get_key(key))
        return self._model.model_validate(value.model_dump()) if value else value

    async def set(self, key: Stringable, value: U | None) -> None:
        await self._store.set(self._get_key(key), value)

    async def watch(self, key: Stringable, *, ready: asyncio.Event | None = None) -> AsyncIterator[U | None]:
        async for value in self._store.watch(self._get_key(key), ready=ready):
            yield self._model.model_validate(value.model_dump()) if value else value

    def _get_key(self, key: Stringable) -> str:
        return f"{self._prefix!s}{key!s}"



import asyncio
import os
from collections.abc import AsyncGenerator, Awaitable
from contextlib import asynccontextmanager
from typing import Any, Callable

import requests
import uvicorn
import uvicorn.config
from fastapi import FastAPI

from acp_sdk.models import Metadata
from acp_sdk.server.agent import Agent
from acp_sdk.server.agent import agent as agent_decorator
from acp_sdk.server.app import create_app
from acp_sdk.server.logging import configure_logger as configure_logger_func
from acp_sdk.server.logging import logger
from acp_sdk.server.store import Store
from acp_sdk.server.telemetry import configure_telemetry as configure_telemetry_func
from acp_sdk.server.utils import async_request_with_retry


class Server:
    def __init__(self) -> None:
        self.agents: list[Agent] = []
        self.server: uvicorn.Server | None = None

    def agent(
        self,
        name: str | None = None,
        description: str | None = None,
        *,
        metadata: Metadata | None = None,
    ) -> Callable:
        """Decorator to register an agent."""

        def decorator(fn: Callable) -> Callable:
            agent = agent_decorator(name=name, description=description, metadata=metadata)(fn)
            self.register(agent)
            return fn

        return decorator

    def register(self, *agents: Agent) -> None:
        self.agents.extend(agents)

    @asynccontextmanager
    async def lifespan(self, app: FastAPI) -> AsyncGenerator[None]:
        yield

    async def serve(
        self,
        *,
        configure_logger: bool = True,
        configure_telemetry: bool = False,
        self_registration: bool = True,
        store: Store | None = None,
        host: str = "127.0.0.1",
        port: int = 8000,
        uds: str | None = None,
        fd: int | None = None,
        loop: uvicorn.config.LoopSetupType = "auto",
        http: type[asyncio.Protocol] | uvicorn.config.HTTPProtocolType = "auto",
        ws: type[asyncio.Protocol] | uvicorn.config.WSProtocolType = "auto",
        ws_max_size: int = 16 * 1024 * 1024,
        ws_max_queue: int = 32,
        ws_ping_interval: float | None = 20.0,
        ws_ping_timeout: float | None = 20.0,
        ws_per_message_deflate: bool = True,
        lifespan: uvicorn.config.LifespanType = "auto",
        env_file: str | os.PathLike[str] | None = None,
        log_config: dict[str, Any]
        | str
        | uvicorn.config.RawConfigParser
        | uvicorn.config.IO[Any]
        | None = uvicorn.config.LOGGING_CONFIG,
        log_level: str | int | None = None,
        access_log: bool = True,
        use_colors: bool | None = None,
        interface: uvicorn.config.InterfaceType = "auto",
        reload: bool = False,
        reload_dirs: list[str] | str | None = None,
        reload_delay: float = 0.25,
        reload_includes: list[str] | str | None = None,
        reload_excludes: list[str] | str | None = None,
        workers: int | None = None,
        proxy_headers: bool = True,
        server_header: bool = True,
        date_header: bool = True,
        forwarded_allow_ips: list[str] | str | None = None,
        root_path: str = "",
        limit_concurrency: int | None = None,
        limit_max_requests: int | None = None,
        backlog: int = 2048,
        timeout_keep_alive: int = 5,
        timeout_notify: int = 30,
        timeout_graceful_shutdown: int | None = None,
        callback_notify: Callable[..., Awaitable[None]] | None = None,
        ssl_keyfile: str | os.PathLike[str] | None = None,
        ssl_certfile: str | os.PathLike[str] | None = None,
        ssl_keyfile_password: str | None = None,
        ssl_version: int = uvicorn.config.SSL_PROTOCOL_VERSION,
        ssl_cert_reqs: int = uvicorn.config.ssl.CERT_NONE,
        ssl_ca_certs: str | None = None,
        ssl_ciphers: str = "TLSv1",
        headers: list[tuple[str, str]] | None = None,
        factory: bool = False,
        h11_max_incomplete_event_size: int | None = None,
    ) -> None:
        if self.server:
            raise RuntimeError("The server is already running")

        if headers is None:
            headers = [("server", "acp")]
        elif not any(k.lower() == "server" for k, _ in headers):
            headers.append(("server", "acp"))

        import uvicorn

        app = create_app(*self.agents, lifespan=self.lifespan, store=store)

        if configure_logger:
            configure_logger_func()
        if configure_telemetry:
            configure_telemetry_func(app)

        config = uvicorn.Config(
            app,
            host,
            port,
            uds,
            fd,
            loop,
            http,
            ws,
            ws_max_size,
            ws_max_queue,
            ws_ping_interval,
            ws_ping_timeout,
            ws_per_message_deflate,
            lifespan,
            env_file,
            log_config,
            log_level,
            access_log,
            use_colors,
            interface,
            reload,
            reload_dirs,
            reload_delay,
            reload_includes,
            reload_excludes,
            workers,
            proxy_headers,
            server_header,
            date_header,
            forwarded_allow_ips,
            root_path,
            limit_concurrency,
            limit_max_requests,
            backlog,
            timeout_keep_alive,
            timeout_notify,
            timeout_graceful_shutdown,
            callback_notify,
            ssl_keyfile,
            ssl_certfile,
            ssl_keyfile_password,
            ssl_version,
            ssl_cert_reqs,
            ssl_ca_certs,
            ssl_ciphers,
            headers,
            factory,
            h11_max_incomplete_event_size,
        )
        self.server = uvicorn.Server(config)
        await self._serve(self_registration=self_registration)

    def run(
        self,
        *,
        configure_logger: bool = True,
        configure_telemetry: bool = False,
        self_registration: bool = True,
        store: Store | None = None,
        host: str = "127.0.0.1",
        port: int = 8000,
        uds: str | None = None,
        fd: int | None = None,
        loop: uvicorn.config.LoopSetupType = "auto",
        http: type[asyncio.Protocol] | uvicorn.config.HTTPProtocolType = "auto",
        ws: type[asyncio.Protocol] | uvicorn.config.WSProtocolType = "auto",
        ws_max_size: int = 16 * 1024 * 1024,
        ws_max_queue: int = 32,
        ws_ping_interval: float | None = 20.0,
        ws_ping_timeout: float | None = 20.0,
        ws_per_message_deflate: bool = True,
        lifespan: uvicorn.config.LifespanType = "auto",
        env_file: str | os.PathLike[str] | None = None,
        log_config: dict[str, Any]
        | str
        | uvicorn.config.RawConfigParser
        | uvicorn.config.IO[Any]
        | None = uvicorn.config.LOGGING_CONFIG,
        log_level: str | int | None = None,
        access_log: bool = True,
        use_colors: bool | None = None,
        interface: uvicorn.config.InterfaceType = "auto",
        reload: bool = False,
        reload_dirs: list[str] | str | None = None,
        reload_delay: float = 0.25,
        reload_includes: list[str] | str | None = None,
        reload_excludes: list[str] | str | None = None,
        workers: int | None = None,
        proxy_headers: bool = True,
        server_header: bool = True,
        date_header: bool = True,
        forwarded_allow_ips: list[str] | str | None = None,
        root_path: str = "",
        limit_concurrency: int | None = None,
        limit_max_requests: int | None = None,
        backlog: int = 2048,
        timeout_keep_alive: int = 5,
        timeout_notify: int = 30,
        timeout_graceful_shutdown: int | None = None,
        callback_notify: Callable[..., Awaitable[None]] | None = None,
        ssl_keyfile: str | os.PathLike[str] | None = None,
        ssl_certfile: str | os.PathLike[str] | None = None,
        ssl_keyfile_password: str | None = None,
        ssl_version: int = uvicorn.config.SSL_PROTOCOL_VERSION,
        ssl_cert_reqs: int = uvicorn.config.ssl.CERT_NONE,
        ssl_ca_certs: str | None = None,
        ssl_ciphers: str = "TLSv1",
        headers: list[tuple[str, str]] | None = None,
        factory: bool = False,
        h11_max_incomplete_event_size: int | None = None,
    ) -> None:
        asyncio.run(
            self.serve(
                configure_logger=configure_logger,
                configure_telemetry=configure_telemetry,
                self_registration=self_registration,
                store=store,
                host=host,
                port=port,
                uds=uds,
                fd=fd,
                loop=loop,
                http=http,
                ws=ws,
                ws_max_size=ws_max_size,
                ws_max_queue=ws_max_queue,
                ws_ping_interval=ws_ping_interval,
                ws_ping_timeout=ws_ping_timeout,
                ws_per_message_deflate=ws_per_message_deflate,
                lifespan=lifespan,
                env_file=env_file,
                log_config=log_config,
                log_level=log_level,
                access_log=access_log,
                use_colors=use_colors,
                interface=interface,
                reload=reload,
                reload_dirs=reload_dirs,
                reload_delay=reload_delay,
                reload_includes=reload_includes,
                reload_excludes=reload_excludes,
                workers=workers,
                proxy_headers=proxy_headers,
                server_header=server_header,
                date_header=date_header,
                forwarded_allow_ips=forwarded_allow_ips,
                root_path=root_path,
                limit_concurrency=limit_concurrency,
                limit_max_requests=limit_max_requests,
                backlog=backlog,
                timeout_keep_alive=timeout_keep_alive,
                timeout_notify=timeout_notify,
                timeout_graceful_shutdown=timeout_graceful_shutdown,
                callback_notify=callback_notify,
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
                ssl_keyfile_password=ssl_keyfile_password,
                ssl_version=ssl_version,
                ssl_cert_reqs=ssl_cert_reqs,
                ssl_ca_certs=ssl_ca_certs,
                ssl_ciphers=ssl_ciphers,
                headers=headers,
                factory=factory,
                h11_max_incomplete_event_size=h11_max_incomplete_event_size,
            )
        )

    async def _serve(self, self_registration: bool = True) -> None:
        registration_task = asyncio.create_task(self._register_agent()) if self_registration else None
        await self.server.serve()
        if registration_task:
            registration_task.cancel()

    @property
    def should_exit(self) -> bool:
        return self.server.should_exit if self.server else False

    @should_exit.setter
    def should_exit(self, value: bool) -> None:
        self.server.should_exit = value

    async def _register_agent(self) -> None:
        """If not in PRODUCTION mode, register agent to the beeai platform and provide missing env variables"""
        if os.getenv("PRODUCTION_MODE", "").lower() in ["true", "1"]:
            logger.debug("Agent is not automatically registered in the production mode.")
            return

        url = os.getenv("PLATFORM_URL", "http://127.0.0.1:8333")
        request_data = {
            "location": f"http://{self.server.config.host}:{self.server.config.port}",
        }
        try:
            await async_request_with_retry(
                lambda client, data=request_data: client.post(f"{url}/api/v1/providers/register/unmanaged", json=data)
            )
            logger.info("Agent registered to the beeai server.")

            # check missing env keyes
            envs_request = await async_request_with_retry(lambda client: client.get(f"{url}/api/v1/variables"))
            envs = envs_request.get("env")
            for agent in self.agents:
                # register all available envs
                missing_keyes = []
                for env in agent.metadata.model_dump().get("env", []):
                    server_env = envs.get(env.get("name"))
                    if server_env:
                        logger.debug(f"Env variable {env['name']} = '{server_env}' added dynamically")
                        os.environ[env["name"]] = server_env
                    elif env.get("required"):
                        missing_keyes.append(env)
                if len(missing_keyes):
                    logger.error(f"Can not run agent, missing required env variables: {missing_keyes}")
                    raise Exception("Missing env variables")

        except requests.exceptions.ConnectionError as e:
            logger.warning(f"Can not reach server, check if running on {url} : {e}")
        except (requests.exceptions.HTTPError, Exception) as e:
            try:
                error_message = e.response.json().get("detail")
                if error_message:
                    logger.warning(f"Agent can not be registered to beeai server: {error_message}")
                else:
                    logger.warning(f"Agent can not be registered to beeai server: {e}")
            except Exception:
                logger.warning(f"Agent can not be registered to beeai server: {e}")



import asyncio
from collections.abc import AsyncGenerator
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager
from datetime import timedelta
from enum import Enum

from fastapi import Depends, FastAPI, HTTPException, status
from fastapi.applications import AppType, Lifespan
from fastapi.encoders import jsonable_encoder
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, StreamingResponse

from acp_sdk.models import (
    Agent as AgentModel,
)
from acp_sdk.models import (
    AgentName,
    AgentReadResponse,
    AgentsListResponse,
    Run,
    RunCancelResponse,
    RunCreateRequest,
    RunCreateResponse,
    RunEventsListResponse,
    RunId,
    RunMode,
    RunReadResponse,
    RunResumeRequest,
    RunResumeResponse,
)
from acp_sdk.models.errors import ACPError
from acp_sdk.models.models import AwaitResume, RunStatus
from acp_sdk.models.schemas import PingResponse
from acp_sdk.server.agent import Agent
from acp_sdk.server.errors import (
    RequestValidationError,
    StarletteHTTPException,
    acp_error_handler,
    catch_all_exception_handler,
    http_exception_handler,
    validation_exception_handler,
)
from acp_sdk.server.executor import CancelData, Executor, RunData
from acp_sdk.server.session import Session
from acp_sdk.server.store import MemoryStore, Store
from acp_sdk.server.utils import stream_sse, wait_util_stop


class Headers(str, Enum):
    RUN_ID = "Run-ID"


def create_app(
    *agents: Agent,
    store: Store | None = None,
    lifespan: Lifespan[AppType] | None = None,
    dependencies: list[Depends] | None = None,
) -> FastAPI:
    executor: ThreadPoolExecutor

    @asynccontextmanager
    async def internal_lifespan(app: FastAPI) -> AsyncGenerator[None]:
        nonlocal executor
        with ThreadPoolExecutor() as exec:
            executor = exec
            if not lifespan:
                yield None
            else:
                async with lifespan(app) as state:
                    yield state

    app = FastAPI(
        lifespan=internal_lifespan,
        dependencies=dependencies,
    )

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["https://agentcommunicationprotocol.dev"],
        allow_methods=["*"],
        allow_headers=["*"],
        allow_credentials=True,
    )

    agents: dict[AgentName, Agent] = {agent.name: agent for agent in agents}

    store = store or MemoryStore(limit=1000, ttl=timedelta(hours=1))
    run_store = store.as_store(model=RunData, prefix="run_")
    run_cancel_store = store.as_store(model=CancelData, prefix="run_cancel_")
    run_resume_store = store.as_store(model=AwaitResume, prefix="run_resume_")
    session_store = store.as_store(model=Session, prefix="session_")

    app.exception_handler(ACPError)(acp_error_handler)
    app.exception_handler(StarletteHTTPException)(http_exception_handler)
    app.exception_handler(RequestValidationError)(validation_exception_handler)
    app.exception_handler(Exception)(catch_all_exception_handler)

    async def find_run_data(run_id: RunId) -> RunData:
        run_data = await run_store.get(run_id)
        if not run_data:
            raise HTTPException(status_code=404, detail=f"Run {run_id} not found")
        if run_data.run.status.is_terminal:
            return run_data
        cancel_data = await run_cancel_store.get(run_data.key)
        if cancel_data is not None:
            run_data.run.status = RunStatus.CANCELLING
        return run_data

    def find_agent(agent_name: AgentName) -> Agent:
        agent = agents.get(agent_name, None)
        if not agent:
            raise HTTPException(status_code=404, detail=f"Agent {agent_name} not found")
        return agent

    @app.get("/agents")
    async def list_agents() -> AgentsListResponse:
        return AgentsListResponse(
            agents=[
                AgentModel(name=agent.name, description=agent.description, metadata=agent.metadata)
                for agent in agents.values()
            ]
        )

    @app.get("/agents/{name}")
    async def read_agent(name: AgentName) -> AgentReadResponse:
        agent = find_agent(name)
        return AgentModel(name=agent.name, description=agent.description, metadata=agent.metadata)

    @app.get("/ping")
    async def ping() -> PingResponse:
        return PingResponse()

    @app.post("/runs")
    async def create_run(request: RunCreateRequest) -> RunCreateResponse:
        agent = find_agent(request.agent_name)

        session = (
            (await session_store.get(request.session_id)) or Session(id=request.session_id)
            if request.session_id
            else Session()
        )
        nonlocal executor
        run_data = RunData(
            run=Run(agent_name=agent.name, session_id=session.id),
            input=request.input,
        )
        await run_store.set(run_data.key, run_data)

        session.append(run_data.run.run_id)
        await session_store.set(session.id, session)

        headers = {Headers.RUN_ID: str(run_data.run.run_id)}
        ready = asyncio.Event()

        Executor(
            agent=agent,
            run_data=run_data,
            history=await session.history(run_store),
            run_store=run_store,
            cancel_store=run_cancel_store,
            resume_store=run_resume_store,
            executor=executor,
        ).execute(wait=ready)

        match request.mode:
            case RunMode.STREAM:
                return StreamingResponse(
                    stream_sse(run_data, run_store, 0, ready=ready),
                    headers=headers,
                    media_type="text/event-stream",
                )
            case RunMode.SYNC:
                await wait_util_stop(run_data, run_store, ready=ready)
                return JSONResponse(
                    headers=headers,
                    content=jsonable_encoder(run_data.run),
                )
            case RunMode.ASYNC:
                ready.set()
                return JSONResponse(
                    status_code=status.HTTP_202_ACCEPTED,
                    headers=headers,
                    content=jsonable_encoder(run_data.run),
                )
            case _:
                raise NotImplementedError()

    @app.get("/runs/{run_id}")
    async def read_run(run_id: RunId) -> RunReadResponse:
        bundle = await find_run_data(run_id)
        return bundle.run

    @app.get("/runs/{run_id}/events")
    async def list_run_events(run_id: RunId) -> RunEventsListResponse:
        bundle = await find_run_data(run_id)
        return RunEventsListResponse(events=bundle.events)

    @app.post("/runs/{run_id}")
    async def resume_run(run_id: RunId, request: RunResumeRequest) -> RunResumeResponse:
        run_data = await find_run_data(run_id)

        if run_data.run.await_request is None:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail=f"Run {run_id} has no await request")

        if run_data.run.await_request.type != request.await_resume.type:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail=f"Run {run_id} is expecting resume of type {run_data.run.await_request.type}",
            )

        run_data.run.status = RunStatus.IN_PROGRESS
        await run_store.set(run_data.key, run_data)
        await run_resume_store.set(run_data.key, request.await_resume)

        match request.mode:
            case RunMode.STREAM:
                return StreamingResponse(
                    stream_sse(run_data, run_store, len(run_data.events)),
                    media_type="text/event-stream",
                )
            case RunMode.SYNC:
                run_data = await wait_util_stop(run_data, run_store)
                return run_data.run
            case RunMode.ASYNC:
                return JSONResponse(
                    status_code=status.HTTP_202_ACCEPTED,
                    content=jsonable_encoder(run_data.run),
                )
            case _:
                raise NotImplementedError()

    @app.post("/runs/{run_id}/cancel")
    async def cancel_run(run_id: RunId) -> RunCancelResponse:
        run_data = await find_run_data(run_id)
        if run_data.run.status.is_terminal:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail=f"Run in terminal status {run_data.run.status} can't be cancelled",
            )
        await run_cancel_store.set(run_data.key, CancelData())
        run_data.run.status = RunStatus.CANCELLING
        return JSONResponse(status_code=status.HTTP_202_ACCEPTED, content=jsonable_encoder(run_data.run))

    return app



import uuid

from pydantic import BaseModel, Field

from acp_sdk.models import Message, RunId, RunStatus, SessionId
from acp_sdk.server.executor import RunData
from acp_sdk.server.store import Store


class Session(BaseModel):
    id: SessionId = Field(default_factory=uuid.uuid4)
    runs: list[RunId] = []

    def append(self, run_id: RunId) -> None:
        self.runs.append(run_id)

    async def history(self, store: Store[RunData]) -> list[Message]:
        history = []
        for run_id in self.runs:
            run_data = await store.get(run_id)
            if run_data is not None and run_data.run.status == RunStatus.COMPLETED:
                history.extend(run_data.input)
                history.extend(run_data.run.output)
        return history



import asyncio
import logging
from collections.abc import AsyncIterator
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from typing import Self

from pydantic import BaseModel, ValidationError

from acp_sdk.instrumentation import get_tracer
from acp_sdk.models import (
    ACPError,
    AnyModel,
    AwaitRequest,
    AwaitResume,
    Error,
    ErrorCode,
    Event,
    GenericEvent,
    Message,
    MessageCompletedEvent,
    MessageCreatedEvent,
    MessagePart,
    MessagePartEvent,
    Run,
    RunAwaitingEvent,
    RunCancelledEvent,
    RunCompletedEvent,
    RunCreatedEvent,
    RunFailedEvent,
    RunInProgressEvent,
    RunStatus,
)
from acp_sdk.server.agent import Agent
from acp_sdk.server.logging import logger
from acp_sdk.server.store import Store


class RunData(BaseModel):
    run: Run
    input: list[Message]
    events: list[Event] = []

    @property
    def key(self) -> str:
        return str(self.run.run_id)

    async def watch(self, store: Store[Self], *, ready: asyncio.Event | None = None) -> AsyncIterator[Self]:
        async for data in store.watch(self.key, ready=ready):
            if data is None:
                raise RuntimeError("Missing data")
            yield data
            if data.run.status.is_terminal:
                break


class CancelData(BaseModel):
    pass


class Executor:
    def __init__(
        self,
        *,
        agent: Agent,
        run_data: RunData,
        history: list[Message],
        executor: ThreadPoolExecutor,
        run_store: Store[RunData],
        cancel_store: Store[CancelData],
        resume_store: Store[AwaitResume],
    ) -> None:
        self.agent = agent
        self.history = history
        self.run_data = run_data
        self.executor = executor

        self.run_store = run_store
        self.cancel_store = cancel_store
        self.resume_store = resume_store

        self.logger = logging.LoggerAdapter(logger, {"run_id": str(run_data.run.run_id)})

    def execute(self, *, wait: asyncio.Event) -> None:
        self.task = asyncio.create_task(self._execute(self.run_data, executor=self.executor, wait=wait))
        self.watcher = asyncio.create_task(self._watch_for_cancellation())

    async def _push(self) -> None:
        await self.run_store.set(self.run_data.run.run_id, self.run_data)

    async def _emit(self, event: Event) -> None:
        freeze = event.model_copy(deep=True)
        self.run_data.events.append(freeze)
        await self._push()

    async def _await(self) -> AwaitResume:
        async for resume in self.resume_store.watch(self.run_data.key):
            if resume is not None:
                await self.resume_store.set(self.run_data.key, None)
                return resume

    async def _watch_for_cancellation(self) -> None:
        while not self.task.done():
            try:
                async for data in self.cancel_store.watch(self.run_data.key):
                    if data is not None:
                        self.task.cancel()
            except Exception:
                logger.warning("Cancellation watcher failed, restarting")

    async def _execute(self, run_data: RunData, *, executor: ThreadPoolExecutor, wait: asyncio.Event) -> None:
        with get_tracer().start_as_current_span("run"):
            in_message = False

            async def flush_message() -> None:
                nonlocal in_message
                if in_message:
                    message = run_data.run.output[-1]
                    message.completed_at = datetime.now(timezone.utc)
                    await self._emit(MessageCompletedEvent(message=message))
                    in_message = False

            try:
                await wait.wait()

                await self._emit(RunCreatedEvent(run=run_data.run))

                generator = self.agent.execute(
                    input=self.history + run_data.input, session_id=run_data.run.session_id, executor=executor
                )
                self.logger.info("Run started")

                run_data.run.status = RunStatus.IN_PROGRESS
                await self._emit(RunInProgressEvent(run=run_data.run))

                await_resume = None
                while True:
                    next = await generator.asend(await_resume)

                    if isinstance(next, (MessagePart, str)):
                        if isinstance(next, str):
                            next = MessagePart(content=next)
                        if not in_message:
                            run_data.run.output.append(Message(parts=[], completed_at=None))
                            in_message = True
                            await self._emit(MessageCreatedEvent(message=run_data.run.output[-1]))
                        run_data.run.output[-1].parts.append(next)
                        await self._emit(MessagePartEvent(part=next))
                    elif isinstance(next, Message):
                        await flush_message()
                        run_data.run.output.append(next)
                        await self._emit(MessageCreatedEvent(message=next))
                        for part in next.parts:
                            await self._emit(MessagePartEvent(part=part))
                        await self._emit(MessageCompletedEvent(message=next))
                    elif isinstance(next, AwaitRequest):
                        run_data.run.await_request = next
                        run_data.run.status = RunStatus.AWAITING
                        await self._emit(RunAwaitingEvent(run=run_data.run))
                        self.logger.info("Run awaited")
                        await_resume = await self._await()
                        run_data.run.status = RunStatus.IN_PROGRESS
                        await self._emit(RunInProgressEvent(run=run_data.run))
                        self.logger.info("Run resumed")
                    elif isinstance(next, Error):
                        raise ACPError(error=next)
                    elif isinstance(next, BaseException):
                        raise next
                    elif next is None:
                        await flush_message()
                    elif isinstance(next, BaseModel):
                        await self._emit(GenericEvent(generic=AnyModel(**next.model_dump())))
                    else:
                        try:
                            generic = AnyModel.model_validate(next)
                            await self._emit(GenericEvent(generic=generic))
                        except ValidationError:
                            raise TypeError("Invalid yield")
            except StopAsyncIteration:
                await flush_message()
                run_data.run.status = RunStatus.COMPLETED
                run_data.run.finished_at = datetime.now(timezone.utc)
                await self._emit(RunCompletedEvent(run=run_data.run))
                self.logger.info("Run completed")
            except asyncio.CancelledError:
                run_data.run.status = RunStatus.CANCELLED
                run_data.run.finished_at = datetime.now(timezone.utc)
                await self._emit(RunCancelledEvent(run=run_data.run))
                self.logger.info("Run cancelled")
            except Exception as e:
                if isinstance(e, ACPError):
                    run_data.run.error = e.error
                else:
                    run_data.run.error = Error(code=ErrorCode.SERVER_ERROR, message=str(e))
                run_data.run.status = RunStatus.FAILED
                run_data.run.finished_at = datetime.now(timezone.utc)
                await self._emit(RunFailedEvent(run=run_data.run))
                self.logger.exception("Run failed")
